// C Source File
// Created 05/29/2008 04:16:12 PM
// Tab size: 5

/*
    Flash Archive Undelete v0.01.01
    Copyright (C) 2008-2009 Travis Evans

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#define COMMENT_PROGRAM_NAME   "Flash Archive Undelete"
#define COMMENT_VERSION_STRING "0.01.01"
#define COMMENT_VERSION_NUMBER 0,0,1,1 /* major, minor, revision, subrevision */
#define COMMENT_AUTHORS        "Travis Evans"



/* Uncomment these lines and recompile to save a few bytes in the binary
   at the expense of requiring AMS 2.04 and up and not being able to transfer
   the program to a calc of a different model */
// #undef MIN_AMS
// #define MIN_AMS 204
// #undef USE_INTERNAL_FLINE_EMULATOR
// #define OPTIMIZE_CALC_CONSTS



#include <tigcclib.h>



/* Calculator model global (optimization so it doesn't have to be
   redetermined over and over again) */
short calculator = 0;



/* String constants */

#define noDeletedVars "No deleted variables found in Flash archive"
#define archiveScanFailed "Sorry, Flash archive scan failed for "\
	" some reason (bug? corrupted archive...?)"
#define insufficientMemory "Insufficient memory"
#define pleaseFreeRAM "Not enough RAM to continue--please free some "\
	"RAM and try again."
#define varMenuTitle "Unarchived vars"
#define scanning "Scanning archive; please wait\xA0"
#define verInfo "Type %s (%u bytes)\x12"
#define undeleteAll "Undelete all\xA0"
#define undeleteAllText1 "Undelete all versions of this variable "
#define undeleteAllText2 "to the following folder:"
#define invalidFolderTitle "Invalid folder name"
#define invalidFolder "'%s' is not allowed as a folder name."
#define folderExists "Folder '%s' exists"
#define useExistingFolder "This folder exists. Use it anyway?"
#define varsRecovered "%d versions were recovered (look for names "\
	"starting with 'var')."
#define varPrefix "var"
#define undelete "Undelete"
#define undeleteTo "Undelete to\xA0"
#define hexView "View hex dump"
#define varExists "Variable exists"
#define replaceVar "Warning: Variable '%s' exists. Replace it with the "\
	"deleted version?"
#define cantReplaceTitle "Cannot replace variable"
#define cantReplace "'%s' is in use and cannot be replaced."
#define hexViewKeyHint "[F1] Toggle hex codes   [ESC] Quit"
#define hexViewKeyHint2 "[\x17] & [\x18] Scroll   "\
	"2nd+[\x17] & 2nd+[\x18] Page"
#define success "Success"
#define undeleteSuccess "'%s' was successfully undeleted."
#define about "About " COMMENT_PROGRAM_NAME
#define about1 COMMENT_PROGRAM_NAME " " COMMENT_VERSION_STRING
#define about2 "Copyright \xA9" "2008-2009 " COMMENT_AUTHORS
#define about3 "This program comes with absolutely no"
#define about4 "warranty. For licensing information,"
#define about5 "see the GNU General Public License at "
#define about6 "http://www.gnu.org/licenses/gpl.html "
#define copyrightHelp "Press [APPS] for copyright & license info"



/* Constants */

#define PATH_STR_LEN 18
#define DLG_MSG_LEN 100
enum PROGRAM_STATUS {UNDELETE_ALL = 4095, SUCCESS = 0, ERROR_RAM = -1,
	USER_CANCEL = -2, VAR_NOT_FOUND = -3, VAR_SAVE_ERROR = -4,
	USER_OTHER_KEY = -5, REFRESH_VAR_MENU = -6, NO_DELETED_VARS = -7,
	ERROR_UNKNOWN = -8};
enum OPERATION_ERROR {UNKNOWN_FAILURE = -1};
enum SECTOR_STATUS {SECTOR_IN_USE = -2, SECTOR_FULL = -4, SECTOR_UNUSED = -1};
enum BLOCK_STATUS {BLOCK_FREE = -1, BLOCK_DELETED = -64, BLOCK_USED = -32};



/* Data structures */

// Struct to keep track of interesting blocks in Archive memory
typedef struct {
	signed short status;     // Block status word
	char path[PATH_STR_LEN]; // Full path + varname with backslash
	char typeStr[5];         // Variable's type string
	ESQ type;                // Variable type
	unsigned short size;     // Variable size
	void *dataStart;         // Pointer to start of data in Flash memory
	                         //     (NOT including size word)
} ARCHIVE_BLOCK_DATA;

// Archive block format
typedef struct {
	signed short status;
	char folder[8];
	char varName[8];
	unsigned short dummy;    // Purpose unknown
	unsigned short next;     // Relative ptr to next block in words
	unsigned short varSize;  // Size of variable data
	char data;               // First byte of actual var data
} ARCHIVE_BLOCK;



//  Functions  //////////////////////////////////////////////////////////////

#define NextFlashSector(ptr) (char *)(((unsigned long)ptr + 0x10000) \
	& 0xFFFF0000)
#define AtSectorStart(ptr) (!((unsigned long)ptr & 0x0000FFFF))
// Thanks to Lionel Debroux for this macro
#define EM_ArchiveMemoryStart() ((!AMS_1xx)?(const unsigned char *) \
	(_rom_call(const unsigned char *,(void),3CF)()): \
	(const unsigned char *)(CALCULATOR?0x590000:0x390000))

/* Fill a ARCHIVE_BLOCK_DATA struct with data about a block in the
   Archive, given a pointer of the first byte of the block header.  If
   the block status is anything other than BLOCK_DELETED or BLOCK_USED,
   all values except status will be set to zero. */
void GetBlockInfo(ARCHIVE_BLOCK *block,	ARCHIVE_BLOCK_DATA *blockData)
{
	ESQ *typeTag = NULL, *type = NULL;
	short varType;

	memset(blockData, 0, sizeof(blockData));
	blockData->status = block->status;

	switch (block->status) {
		case BLOCK_DELETED:
		case BLOCK_USED:
			strncat(blockData->path, block->folder, 8);
			strcat(blockData->path, "\\");
			strncat(blockData->path, block->varName, 8);
			blockData->size = block->varSize;
			blockData->dataStart = &block->data;

			/* Get type string */
			typeTag = (char *)
				((unsigned long)(&block->data) +
				block->varSize - 1);
			varType = GetDataType(typeTag);
			blockData->type = *typeTag;
			if (varType != SDT_OTH) {
				strncpy(blockData->typeStr,
					SmapTypeStrings(varType), 5);
			} else {
				// Find custom file extension
				type = typeTag - sizeof(short);
				while (!(*type == 0)) type--;
				type++;
				strncpy(blockData->typeStr, type, 5);
			}
	}

}

/* Given a pointer to the start of an Archive block header, return a pointer
   to the beginning of the next block (or NULL if there are no more blocks
   in the current sector) */
ARCHIVE_BLOCK *GetNextBlock(ARCHIVE_BLOCK *block)
{
	unsigned short blockSizeBytes;
	ARCHIVE_BLOCK *nextBlock;

	if (block->status == BLOCK_FREE) return NULL;
	blockSizeBytes = block->next << 1;
	nextBlock = (ARCHIVE_BLOCK *)(
		(unsigned long)(&(block->next)) + blockSizeBytes);
	if (AtSectorStart(nextBlock)) return NULL;
	return nextBlock;
}

/* Scans Archive memory for deleted/unarchived variables, filling
   delVars with an array of ARCHIVE_BLOCK_DATA structs for each one (if
   not NULL).  Returns the number of deleted variables found on success (may
   be zero) or an OPERATION_ERROR code on failure. */
signed long DoScan(ARCHIVE_BLOCK_DATA *delVars)
{
	void *ptr = NULL, *next = NULL;
	ARCHIVE_BLOCK_DATA blockData;
	signed short sectorStatus;
	unsigned short blocksScanned = 0, nDelVars = 0;
	unsigned char needNextSector = FALSE;

	/* Scan Archive for deleted vars */
	ptr = (void *)(unsigned char *)EM_ArchiveMemoryStart();
	if (!ptr) return UNKNOWN_FAILURE;
	while ((unsigned char *)ptr < FlashMemoryEnd) {

		// Get sector status word which begins a sector
		sectorStatus = *(signed short *)ptr;
		// Advance to first byte of block header which follows
		((signed short *)ptr)++;

		if (sectorStatus != SECTOR_UNUSED) {

			/* Continue checking blocks until either we find a
			   free block or we find ourselves sitting at the very
			   beginning of the next sector.  In either case, it's
			   time to reread the next sector status flag. */
			needNextSector = FALSE;
			while (!needNextSector) {
				// Even address sanity-check (why not?)
				if ((unsigned long)ptr & 0x00000001) {
					return UNKNOWN_FAILURE;
				}
				GetBlockInfo((ARCHIVE_BLOCK *)ptr,	&blockData);
				blocksScanned++;
				/* Not only check if the block was deleted but that the
				   variable size is actually set--for some weird reason,
				   my calc actually had bogus blocks which had all the
				   right flags set but were otherwise totally empty--it
				   caused the program to put some garbage entries in the
				   variable list */
				if (blockData.status == BLOCK_DELETED &&
					((ARCHIVE_BLOCK *)ptr)->varSize != 0xFFFF) {

					if (delVars) {
						memcpy(&delVars[nDelVars], &blockData,
							sizeof(ARCHIVE_BLOCK_DATA));
					}
					nDelVars++;
				} else if (blockData.status == BLOCK_FREE) {
					needNextSector = TRUE;
				}
					next = GetNextBlock((ARCHIVE_BLOCK *)ptr);
				if (next != NULL) {
					ptr = next;
				} else {
					needNextSector = TRUE;
				}

			}

		}

		if (!AtSectorStart(ptr))
			ptr = NextFlashSector(ptr);

	}

	return nDelVars;
}

/* Undeletes a variable given the ARCHIVE_BLOCK_DATA struct and an HSym to a
   newly created variable to undelete to. */
signed char Undelete(ARCHIVE_BLOCK_DATA *var, HSym hs)
{
	HANDLE h;
	void *src = NULL, *dest = NULL;

	if (!hs.folder) return ERROR_RAM;
	h = HeapAlloc(var->size + sizeof(var->size));
	if (h == H_NULL) return ERROR_RAM;
	HeapLock(h);
	src = var->dataStart;
	dest = HeapDeref(h);
	*(short *)dest = var->size;
	((short *)dest)++;
	memcpy(dest, src, var->size);
	HeapUnlock(h);
	DerefSym(hs)->handle = h;
	return SUCCESS;
}

//  UI Functions ////////////////////////////////////////////////////////////

/* Additional keycodes */
#define KEY_2ND 4096

/* Screen coordinates of text and display elements */
#define TOPLINE89_Y 1
#define TOPLINE92P_Y 1
#define BOTLINE89_Y 100 - 14      // 100 = LCD height on TI-89(t)
#define BOTLINE92P_Y 128 - 15     // 128 = LCD height on TI-92+/V200
#define VERYBOTLINE89_Y 100 - 5
#define VERYBOTLINE92P_Y 128 - 5
#define STAT_LINE_BORDER_Y LCD_HEIGHT - 7

/* Clear entire screen, then redraw status line border */
void Clear(void)
{
	ClrScr();
	DrawLine(0, STAT_LINE_BORDER_Y, calculator?239:159,
		STAT_LINE_BORDER_Y, A_NORMAL);
}

/* Wait until a key is pressed, idling the CPU in low power mode in the
   meantime.  Return keycode from keyboard queue. */
unsigned short KeyIdle()
{
		unsigned short key;
		void *kbq = kbd_queue();

		OSTimerRestart(APD_TIMER);
		while (OSdequeue(&key, kbq)) {
			if (OSTimerExpired(APD_TIMER)) {
				off();
				OSTimerRestart(APD_TIMER);
			}
			idle();
		}

		ST_eraseHelp();
		key &= 0xF7FF;

		return key;
}

/* Displays a menu of variables given an array of ARCHIVE_BLOCK_DATA and
   the number of elements it contains.  If successful (>=0 return value),
   sets pathStr to the full path of the variable selected.

   IMPORTANT:  VarMenuDestroy() (which actually calls VarMenu() with null
   parameters) must be called after the last use of VarMenu() to avoid a
   memory leak.  The allocated menu is left open to make subsequent menu
   displays instant (it can take a long time to create/sort the menu when
   there are many variables to list). */
signed short VarMenu(ARCHIVE_BLOCK_DATA *vars,
	signed short nVars, char *pathStr)
{
	static HANDLE menu = H_NULL;
	static char (*varList)[PATH_STR_LEN] = NULL;
	unsigned char found = FALSE;
	signed short i, j, nVarList = 0;
	signed short selection = ERROR_RAM;
	char menuStr[19];

	/* Clean up instead of displaying a menu? */
	if (vars == NULL) {
		if (menu != H_NULL) HeapFree(menu);
		menu = H_NULL;
		if (varList != NULL) free(varList);
		varList = NULL;
		return SUCCESS;
	}

	/* See if we need to populate the menu; skip otherwise */
	if (menu == H_NULL) {

		/* Make a sorted list of variables without duplicates */
		varList = calloc(nVars, PATH_STR_LEN * sizeof(char));
		if (!varList) goto Cleanup;
		for (i = 0; i < nVars; i++) {
			j = 0;
			found = FALSE;
			while (j < nVarList) {
				if (!strcmp(vars[i].path, varList[j])) {
					found = TRUE;
					break;
				}
				j++;
			}
			if (!found) {
				strcpy(varList[nVarList], vars[i].path);
				nVarList++;
			}
		}
		qsort(varList[0], nVarList, PATH_STR_LEN * sizeof(char),
			(compare_t)strcmp);

		/* Populate menu */
		menu = PopupNew(varMenuTitle, 0);
		if (!menu) goto Cleanup;
		for (i = 0; i < nVarList; i++) {
			strcpy(menuStr, varList[i]);
			strcat(menuStr, "\x12");
			if (!PopupAddText(menu, -1, menuStr, 0)) goto Cleanup;
		}

	}

	/* Get input and return */
	selection = PopupDo(menu, CENTER, CENTER, 0);
	if (!selection) {
		/* If user pressed a key like APPS, CATALOG, etc., the keypress
		   will be stuck in the key buffer, so let the caller know about
		   it; otherwise, assume ESC (cancel) */
		if (kbhit()) {
			selection = USER_OTHER_KEY;
		} else {
			selection = USER_CANCEL;
		}
	} else {
		selection--;
		strcpy(pathStr, varList[selection]);
	}

Cleanup:
	return selection;
}

/* Cleans up dynamic memory created by VarMenu().  MUST be called some
   time before program exit or a memory/handle leak will result.  Returns
   SUCCESS. */
inline signed short VarMenuDestroy(void) {
	return VarMenu(NULL, 0, NULL);
}

/* Displays a menu of versions of a particular variable given its path.
   Returns a subscript to the element in vars for the selected item on
   success, UNDELETE_ALL if the user wants to undelete all versions to a
   folder, or a PROGRAM_STATUS error code on failure. */
signed short VersionMenu(ARCHIVE_BLOCK_DATA *vars, signed short nVars,
	char *pathStr)
{

	HANDLE menu = H_NULL;
	char menuStr[25];
	signed short i, j = 0, selection = ERROR_RAM, nVers = 0;

	menu = PopupNew(pathStr, 0);
	if (!menu) goto Cleanup;
	for (i = 0; i < nVars; i++) {
		if (!strcmp(pathStr, vars[i].path)) {
			sprintf(menuStr, verInfo, vars[i].typeStr,
				vars[i].size);
			if (!PopupAddText(menu, -1, menuStr, 0)) goto Cleanup;
			nVers++;
		}
	}
	if (nVers > 1) {
		if (!PopupAddText(menu, -1, undeleteAll, UNDELETE_ALL))
			goto Cleanup;
	} else if (!nVers) {
		selection = VAR_NOT_FOUND;
		goto Cleanup;
	}

	selection = PopupDo(menu, CENTER, CENTER, 0);
	if (!selection) {
		/* If user pressed a key like APPS, CATALOG, etc., the keypress
		   will be stuck in the key buffer, so let the caller know about
		   it; otherwise, assume ESC (cancel) */
		if (kbhit()) {
			selection = USER_OTHER_KEY;
		} else {
			selection = USER_CANCEL;
		}
	} else if (selection != UNDELETE_ALL) {
		// Find selected item in vars
		i = selection;
		j = 0;
		while (i > 0) {
			if (!strcmp(pathStr, vars[j].path)) i--;
			j++;
		}
		selection = j - 1;
	}

Cleanup:
	if (menu != H_NULL) HeapFree(menu);
	return selection;
}

/* Undelete a variable to its original location (and ask the user first if
   the variable already exists) */
signed char UndeleteNow(ARCHIVE_BLOCK_DATA *var)
{
	unsigned char i;
	char dlgMsg[DLG_MSG_LEN];
	char folderName[9];
	signed short status = SUCCESS;
	HSym newVar;
	SYM_ENTRY *newVarEntry;

	newVar = SymFind(SYMSTR(var->path));
	if (newVar.folder) {
		sprintf(dlgMsg, replaceVar, var->path);
		if (DlgMessage(varExists, dlgMsg, BT_YES, BT_NO) ==
			KEY_ESC) return USER_CANCEL;
		newVarEntry = DerefSym(newVar);
		if (newVarEntry->flags.bits.hidden ||
			newVarEntry->flags.bits.twin) {

			sprintf(dlgMsg, cantReplace, var->path);
			DlgMessage(cantReplaceTitle, dlgMsg, BT_OK,
				BT_NONE);
			return USER_CANCEL;
		} else if (newVarEntry->flags.bits.archived) {
			EM_moveSymFromExtMem(NULL, newVar);
			status = REFRESH_VAR_MENU;
		}
		SymDel(SYMSTR(var->path));
	}

	/* Create original folder if it doesn't exist anymore */
	for (i = 0; var->path[i] != '\\'; i++) {
		folderName[i] = var->path[i];
	}
	folderName[i] = 0;
	if (FolderFind(SYMSTR(folderName)) == NOT_FOUND) {
		FolderAdd(SYMSTR(folderName));
	}

	newVar = SymAdd(SYMSTR(var->path));
	if (Undelete(var, newVar) == ERROR_RAM) {
		SymDel(SYMSTR(var->path));
		return ERROR_RAM;
	}
	sprintf(dlgMsg, undeleteSuccess, var->path);
	DlgMessage(success, dlgMsg, BT_OK, BT_NONE);
	return status;
}

/* Displays the dialog box for undeleting a variable and performs the
   undeletion if desired.  var is the ARCHIVE_BLOCK_DATA struct for the
   var to undelete. */
signed char UndeleteDialog(ARCHIVE_BLOCK_DATA *var)
{
	ESQ varType[2] = {OTH_TAG, 0};
	const char *typeStrArray[2] = {NULL, NULL};
	char typeStr[5], dlgMsg[DLG_MSG_LEN];
	signed char result;
	HSym newVar;

	/* For some reason VarNew doesn't display the type to be saved as
	   expected using many variables' native tags, so we always force
	   the type string to what we want by using OTH_TAG (the Undelete
	   function will write the proper tag when the variable is
	   undeleted). */
	strncpy(typeStr, var->typeStr, 5);
	typeStrArray[0] = typeStr;
	newVar = VarNew(varType, typeStrArray);
	if (newVar.folder == H_NULL) {
		return USER_CANCEL;
	}

	result = Undelete(var, newVar);
	if (result == SUCCESS) {
		sprintf(dlgMsg, undeleteSuccess, var->path);
		DlgMessage(success, dlgMsg, BT_OK, BT_NONE);
	}

	return result;
}

/* Displays a simple hex viewer for the selected deleted var. */
signed short HexView(ARCHIVE_BLOCK_DATA *var)
{
	const short CHAR_W  = 6, CHAR_H  = 8, MARGIN89_X = 2, MARGIN89_Y = 11,
		MARGIN92P_X = 0, MARGIN92P_Y = 9;

	unsigned char hexCodes = TRUE, hexLen, charLen, dispLines;
	unsigned char needDispUpdate = TRUE, newDispLayout = TRUE;
	unsigned short charsPerScr;
	unsigned char i, j, x, y;
	unsigned long top = 0, offset = 0;
	unsigned short key;
	char str[5];

	/* Screen setup */
	dispLines = calculator ? 14 : 10;

	while (TRUE) {
		/* Update screen */
		if (newDispLayout) {
			Clear();
			DrawLine(0, 7, calculator ? 239 : 159, 7, A_NORMAL);
			x = calculator ? MARGIN92P_X : MARGIN89_X;
			x += CHAR_W * 4 + 3, 7;
			DrawLine(x, 7, x, STAT_LINE_BORDER_Y, A_NORMAL);
			FontSetSys(F_4x6);
			DrawStr(0, calculator ? TOPLINE92P_Y : TOPLINE89_Y,
				hexViewKeyHint, A_NORMAL);
			ST_busy(ACTIVITY_IDLE);
			FontSetSys(F_6x8);
			needDispUpdate = TRUE;
			newDispLayout = FALSE;
		}

		if (needDispUpdate) {
			offset = top;
			y = calculator ? MARGIN92P_Y : MARGIN89_Y;
			if (hexCodes) {
				charLen = calculator ? 10 : 6;
			} else {
				charLen = calculator ? 35 : 21;
			}
			hexLen = hexCodes ? charLen : 0;
			charsPerScr = charLen * dispLines;

			for (i = 0; i < dispLines; i++) {
				x = calculator ? MARGIN92P_X : MARGIN89_X;
				if (offset < var->size) {
					sprintf(str, "%04lX", offset);
					DrawStr(x, y, str, A_REPLACE);
				} else {
					DrawStr(x, y, "    ", A_REPLACE);
				}
				x += CHAR_W * 5;
				for (j = 0; j < hexLen; j++) {
					if (offset < var->size) {
						sprintf(str, "%02X", ((unsigned char *)(var->
							dataStart))[offset]);
						DrawStr(x, y, str, A_REPLACE);
					} else {
						DrawStr(x, y, "  ", A_REPLACE);
					}
					x += CHAR_W * (2 + j % 2);
					offset++;
				}
				offset -= hexLen;
				for (j = 0; j < charLen; j++) {
					if (offset < var->size) {
						DrawChar(x, y,	((unsigned char *)(var->
							dataStart))[offset], A_REPLACE);
					} else {
						DrawChar(x, y, ' ', A_REPLACE);
					}
					x += CHAR_W;
					offset++;
				}
				y += CHAR_H;
			}

			needDispUpdate = FALSE;
		}

		ST_helpMsg(hexViewKeyHint2);
		key = KeyIdle();
		needDispUpdate = TRUE;

		if (key == KEY_F1) {
			hexCodes = !hexCodes;
			newDispLayout = TRUE;
		} else if (key == KEY_ESC) {
			break;
		} else if (key == KEY_RIGHT || key == KEY_DOWN + KEY_2ND) {
			if (top + charsPerScr < var->size) {
				top += charsPerScr;
			}
		} else if (key == KEY_LEFT || key == KEY_UP + KEY_2ND) {
			if (top > charsPerScr) {
				top -= charsPerScr;
			} else {
				top = 0;
			}
		} else if (key == KEY_UP) {
			if (top > charLen) {
				top -= charLen;
			} else {
				top = 0;
			}
		} else if (key == KEY_DOWN) {
			if (top + charLen < var->size) {
				top += charLen;
			}
		} else if (key == (unsigned)KEY_DIAMOND + KEY_UP) {
			top = 0;
		} else if (key == (unsigned)KEY_DIAMOND + KEY_DOWN) {
			if (var->size > charsPerScr) {
				top = var->size - charsPerScr;
			}
		} else {
			// Unrecognized key; do nothing
			needDispUpdate = FALSE;
		}
	}

	Clear();
	return USER_CANCEL;
}

/* Displays the menu of operations to perform on the selected deleted var.
    */
signed short OpMenu(ARCHIVE_BLOCK_DATA *var)
{
	enum MENU_CHOICES {UNDELETE = 1, UNDELETE_TO = 2, HEX_VIEW = 3};

	HANDLE menu = H_NULL;
	signed short result = ERROR_RAM;

	menu = PopupNew(var->path, 0);
	if (!menu) goto Cleanup;
	PopupAddText(menu, -1, undelete, UNDELETE);
	PopupAddText(menu, -1, undeleteTo, UNDELETE_TO);
	PopupAddText(menu, -1, hexView, HEX_VIEW);

	while (TRUE) {
		result = PopupDo(menu, CENTER, CENTER, 0);
		if (result == 0) {
			/* If user pressed a key like APPS, CATALOG, etc., the
			   keypress will be stuck in the key buffer, so let the
			   caller know about it; otherwise, assume ESC (cancel) */
			if (kbhit()) {
				result = USER_OTHER_KEY;
			} else {
				result = USER_CANCEL;
			}
			break;
		} else if (result == UNDELETE) {
			result = UndeleteNow(var);
		} else if (result == UNDELETE_TO) {
			result = UndeleteDialog(var);
		} else if (result == HEX_VIEW) {
			result = HexView(var);
		}

		if (result == SUCCESS || result == REFRESH_VAR_MENU) break;
		if (result == ERROR_RAM)
			DlgMessage(insufficientMemory, pleaseFreeRAM, BT_OK, BT_NONE);
	}

Cleanup:
	if (menu != H_NULL) HeapFree(menu);
	return result;
}

/* Displays the dialog box for undeleting all variables of the same path to a
   folder.  vars is the ARCHIVE_BLOCK_DATA array to search through. */
signed char UndeleteAllDialog(char *path, ARCHIVE_BLOCK_DATA *vars,
	short nVars)
{
	HANDLE dlg = H_NULL;
	HSym newVar = HS_NULL;
	char folderName[9], newPath[PATH_STR_LEN], dlgMsg[DLG_MSG_LEN];
	short result = ERROR_RAM, i, varNo = 0, varsUndel = 0;

	folderName[0] = 0;
	dlg = DialogNewSimple(0, 0);
	if (!dlg) goto Cleanup;
	DialogAddTitle(dlg, path, BT_OK, BT_CANCEL);
	DialogAddText(dlg, 2, 14, undeleteAllText1);
	DialogAddText(dlg, 2, 22, undeleteAllText2);
	DialogAddRequest(dlg, 10, 32, "", 0, 8, 10);

	while (TRUE) {
		if (DialogDo(dlg, CENTER, CENTER, folderName, NULL) ==
			KEY_ENTER) {

			/* Create the folder if necessary */
			result = FolderFind(SYMSTR(folderName));
			if (result == NOT_FOUND) {
				if (ValidateSymName(folderName)) {
					FolderAdd(SYMSTR(folderName));
				} else {
					sprintf(dlgMsg, invalidFolder,
						folderName);
					DlgMessage(invalidFolderTitle, dlgMsg,
						BT_OK, BT_NONE);
					break;
				}
			} else if (result == FOLDER_TABLE ||
				result == MAIN_FOLDER) {
				sprintf(dlgMsg, folderExists, folderName);
				if (DlgMessage(dlgMsg, useExistingFolder,
					BT_YES, BT_NO) == KEY_ESC) {
					break;
				}
			}

			/* Undelete the versions */
			i = 0;
			while (i < nVars) {
				if (!strcmp(path, vars[i].path)) {
					sprintf(newPath, "%s\\%s%05d",
						folderName, varPrefix, varNo);
					if (!SymFind(SYMSTR(newPath)).folder) {
						newVar = SymAdd(SYMSTR(newPath));
						result = Undelete(&vars[i],
							newVar);
						if (result == ERROR_RAM)
							goto Cleanup;
						i++;
						varsUndel++;
					}
					varNo++;
				} else {
					i++;
				}
			}

			result = SUCCESS;
			break;
		} else {
			result = USER_CANCEL;
			break;
		}
	}

Cleanup:
	if (dlg != H_NULL) HeapFree(dlg);
	if (result == ERROR_RAM) {
		SymDel(SYMSTR(newPath));
	} else if (result == SUCCESS) {
		sprintf(dlgMsg, varsRecovered, varsUndel);
		DlgMessage(path, dlgMsg, BT_OK, BT_NONE);
	}
	return result;
}

/* Screen coordinates of program ID strings */
#define PROG189_X 25
#define PROG289_X 11
#define PROG192P_X 65
#define PROG292P_X 51

/* Displays program identification text at the top and bottom of the screen
   */
void DispProgID(void)
{
	short str1_x, str1_y, str2_x, str2_y;

	FontSetSys(F_4x6);
	Clear();

	if (calculator) {
		str1_x = PROG192P_X;
		str1_y = TOPLINE92P_Y;
		str2_x = PROG292P_X;
		str2_y = BOTLINE92P_Y;
	} else {
		str1_x = PROG189_X;
		str1_y = TOPLINE89_Y;
		str2_x = PROG289_X;
		str2_y = BOTLINE89_Y;
	}
	DrawStr(str1_x, str1_y, COMMENT_PROGRAM_NAME " v"
		COMMENT_VERSION_STRING, A_NORMAL);
	DrawStr(str2_x, str2_y, copyrightHelp, A_NORMAL);
}

/* Displays license and copyright dialog */
void Copyright(void) {
	HANDLE dlg = H_NULL;

	Clear();
	dlg = DialogNewSimple(0, 0);
	if (!dlg) {
		DlgMessage(insufficientMemory, pleaseFreeRAM, BT_OK, BT_NONE);
		return;
	}
	DialogAddTitle(dlg, about, BT_OK, BT_NONE);
	DialogAddText(dlg, 2, 14, about1);
	DialogAddText(dlg, 2, 22, about2);
	DialogAddText(dlg, 2, 30, about3);
	DialogAddText(dlg, 2, 38, about4);
	DialogAddText(dlg, 2, 46, about5);
	DialogAddText(dlg, 2, 54, about6);
	DialogDo(dlg, CENTER, CENTER, NULL, NULL);
	HeapFree(dlg);
}

/* Runs the main UI for the program.  delVars must be a pointer to a
   list of ARCHIVE_BLOCK_DATA structs and nDelVars the number of items
   in the struct array. Returns SUCCESS unless some type of fatal error
   occurred. */
signed char DoMainUI(ARCHIVE_BLOCK_DATA *vars, signed short nVars)
{
	enum CONTEXTS {VAR_MENU, VER_MENU, OP_MENU, UNDEL_DLG,
		UNDEL_ALL_DLG, QUIT};

	unsigned char context = VAR_MENU;
	signed short result = 0, varNo = 0;
	char path[PATH_STR_LEN];

	while (context != QUIT) {
		switch (context) {
			case VAR_MENU:
				// Get a variable to operate on
				DispProgID();
				result = VarMenu(vars, nVars, path);
				if (result == USER_OTHER_KEY) {
					if (ngetchx() == KEY_APPS) {
						// License/copyright dialog
						Copyright();
					}
					break;
				} else if (result < 0) {
					context = QUIT;
					break;
				}
				context++;
			case VER_MENU:
				// Get a version of that variable
				Clear();
				result = VersionMenu(vars, nVars, path);
				if (result == USER_CANCEL) {
					context--;
					break;
				} else if (result == USER_OTHER_KEY) {
					ngetchx();
					break;
				} else if (result == UNDELETE_ALL) {
					context = UNDEL_ALL_DLG;
					break;
				} else if (result < 0) {
					context = QUIT;
					break;
				}
				varNo = result;
				context++;
			case OP_MENU:
				// Operation menu
				Clear();
				result = OpMenu(&vars[varNo]);
				if (result == USER_CANCEL) {
					context--;
				} else if (result == USER_OTHER_KEY) {
					ngetchx();
					break;
				} else if (result == SUCCESS) {
					context = VAR_MENU;
				} else if (result == REFRESH_VAR_MENU) {
					context = QUIT;
					break;
				} else if (result == ERROR_RAM) {
					DlgMessage(insufficientMemory, pleaseFreeRAM,
						BT_OK, BT_NONE);
					context--;
				}
				break;
			case UNDEL_ALL_DLG:
				// Undelete all versions dialog
				result = UndeleteAllDialog(path, vars, nVars);
				if (result == USER_CANCEL) {
					context = VER_MENU;
				} else if (result == SUCCESS) {
					context = VAR_MENU;
				} else if (result == ERROR_RAM) {
					DlgMessage(insufficientMemory, pleaseFreeRAM,
						BT_OK, BT_NONE);
					context = VER_MENU;
				}
		}
	}

	VarMenuDestroy();
	return result;

}

void _main(void)
{
	signed short nDelVars, status = REFRESH_VAR_MENU;
	ARCHIVE_BLOCK_DATA *delVars;

	calculator = CALCULATOR;

	/* Collect data about all deleted variables */
	while (status == REFRESH_VAR_MENU) {
		ST_helpMsg(scanning);
		nDelVars = DoScan(NULL);
		if (nDelVars > 0) {
			delVars = (ARCHIVE_BLOCK_DATA *)
				calloc(nDelVars, sizeof(ARCHIVE_BLOCK_DATA));
			DoScan(delVars);
			if (delVars) {
				status = DoMainUI(delVars, nDelVars);
				free(delVars);
			} else {
				status = ERROR_RAM;
			}
			ST_eraseHelp();
			if (status == ERROR_RAM) {
				DlgMessage(insufficientMemory, pleaseFreeRAM, BT_OK, BT_NONE);
			}
		} else if (nDelVars == 0) {
			status = NO_DELETED_VARS;
			ST_eraseHelp();
			DlgMessage(COMMENT_PROGRAM_NAME, noDeletedVars, BT_OK, BT_NONE);
		} else {
			status = ERROR_UNKNOWN;
			ST_eraseHelp();
			DlgMessage(COMMENT_PROGRAM_NAME, archiveScanFailed, BT_OK,
				BT_NONE);
		}
	}
}


